#ifndef MATRIX_H
#define MATRIX_H

class Vector;
class Point;
class Quaternion;

class Matrix
{
	// Row major order
	float _11; float _12; float _13; float _14;
	float _21; float _22; float _23; float _24;
	float _31; float _32; float _33; float _34;
	float _41; float _42; float _43; float _44;

public:
	Matrix();
	Matrix(const int i);
	Matrix(const float a[16]);
	Matrix(const float a[4][4]);
	Matrix(float A_11, float A_12, float A_13,
	       float A_21, float A_22, float A_23,
	       float A_31, float A_32, float A_33);
	Matrix(float A_11, float A_12, float A_13, float A_14,
	       float A_21, float A_22, float A_23, float A_24,
	       float A_31, float A_32, float A_33, float A_34,
	       float A_41, float A_42, float A_43, float A_44);
	Matrix(const Vector &v1, const Vector &v2, const Vector &v3);   // Column vectors

	Matrix &operator=(const Matrix &B);

	friend Matrix diag(float A_11, float A_22, float A_33, float A_44);

	operator float*();

	Matrix operator+() const;
	Matrix operator-() const;

	Matrix operator!() const;   // Inverse
	Matrix operator~() const;   // Transpose

	Matrix &operator+=(const Matrix &B);
	Matrix &operator-=(const Matrix &B);
	Matrix &operator*=(float s);
	Matrix &operator*=(const Matrix &B);
	Matrix &operator/=(float s);

	float *operator[](int i);          // Access element [row, col], starting with [0][0]
	float &operator()(int i, int j);   // Access element (row, col), starting with (1, 1)

	const float *operator[](int i) const;          // Access elements [row, col] starting with [0][0]
	const float &operator()(int i, int j) const;   // Access elements (row, col) starting with (1, 1)

	friend bool operator==(const Matrix &A, const Matrix &B);
	friend bool operator!=(const Matrix &A, const Matrix &B);

	friend Matrix operator+(const Matrix &A, const Matrix &B);
	friend Matrix operator-(const Matrix &A, const Matrix &B);
	friend Matrix operator*(float s, const Matrix &A);
	friend Matrix operator*(const Matrix &A, const Matrix &B);
	friend Matrix operator/(const Matrix &A, float s);
	friend Matrix operator^(const Matrix &A, const int n);

	Vector operator*(const Vector& v) const;
	Point operator*(const Point& P) const;
	Quaternion operator*(const Quaternion& Q) const;

	friend Vector operator*(const Vector &v, const Matrix &A);
	friend Point operator*(const Point &P, const Matrix &A);
	friend Quaternion operator*(const Quaternion &Q, const Matrix &A);

	friend Vector &operator*=(Vector &v, const Matrix &A);
	friend Point &operator*=(Point &P, const Matrix &A);
	friend Quaternion &operator*=(Quaternion &Q, const Matrix &A);

	friend float det(const Matrix &A);
	friend float det(float A_11);
	friend float det(float A_11, float A_12,
	                 float A_21, float A_22);
	friend float det(float A_11, float A_12, float A_13,
	                 float A_21, float A_22, float A_23,
	                 float A_31, float A_32, float A_33);
	friend float det(float A_11, float A_12, float A_13, float A_14,
	                 float A_21, float A_22, float A_23, float A_24,
	                 float A_31, float A_32, float A_33, float A_34,
	                 float A_41, float A_42, float A_43, float A_44);
	friend float det(const Vector &v1, const Vector &v2, const Vector &v3);

	friend float tr(const Matrix &A);
};

inline float *Matrix::operator[](int i)
{
	return (float*)(&_11 + (i << 2));
}

inline float &Matrix::operator()(int i, int j)
{
	return *((float*)(&_11 + (((i - 1) & 3) << 2) + ((j - 1) & 3)));
}

inline const float *Matrix::operator[](int i) const
{
	return (float*)(&_11 + (i << 2));
}

inline const float &Matrix::operator()(int i, int j) const
{
	return *((float*)(&_11 + (((i - 1) & 3) << 2) + ((j - 1) & 3)));
}

inline Matrix operator^(const Matrix &A, const int n)
{
	// Switch will be resolved at compile time!

	switch(n)
	{
		case -1: return !A;
		case  0: return 0;
		case  1: return A;
		case  2: return A*A;
		default: return A;
	}
}

#endif   // MATRIX_H